#include "StdAfx.h"
#include "StorageImpl.h"
#include <atlbase.h>

CStorageImpl::CStorageImpl(void)
{
	CoInitialize(NULL);
	pHeader=NULL;
}

CStorageImpl::~CStorageImpl(void)
{
	char szDebug[128];
	wsprintf(szDebug,"DestroyStorageImpl 0x%x",this);
	OutputDebugString(szDebug);
	ClearAllItems();
}

void CStorageImpl::InitializeForForstRun(){
//	AddItem("Full Control",TRUE,NULL);
//	AddItem("View Only",TRUE,NULL);
}

void CStorageImpl::InsertToRoot(PSTORAGE_INFO pNew){
	pNew->pNext =pHeader;
	if(pHeader){
		pHeader->pPrev=pNew;
	}
	pHeader=pNew;
}

void CStorageImpl::InsertToFirst(PSTORAGE_INFO pNew,PSTORAGE_INFO pParent){
	pNew->pNext =pParent->pSub;
	if(pParent->pSub){
		pParent->pSub->pPrev=pNew;
	}
	pParent->pSub=pNew;
	pNew->pParent =pParent;

}

PSTORAGE_INFO CStorageImpl::AddItem(const char *pszName,BOOL bFolder,PSTORAGE_INFO pParent){
	PSTORAGE_INFO pNew;

	LONG lSize;
	lSize=(long)sizeof(STORAGE_INFO)+(long)strlen(pszName)+4;
	lSize>>=2;
	lSize<<=2;
	pNew=(PSTORAGE_INFO)new char[lSize];
	memset((void*)pNew,0,lSize);

	pNew->bFolder =bFolder;
	pNew->identifier =MY_FOLDER_IDENTIFIER;
	pNew->cb=(USHORT)lSize-sizeof(pNew->cb);
	strcpy(pNew->szName,pszName);
	if(pParent){
		InsertToFirst(pNew,pParent);
	}else{
		InsertToRoot(pNew);
	}

	char szDebug[128];
	wsprintf(szDebug,"0x %x NewItem %s",pNew,pNew->szName);
	OutputDebugString(szDebug);
	return pNew;
}

void CStorageImpl::SkipCurrentElement(PSTORAGE_INFO pItem){
	if(pItem->pPrev){
		pItem->pPrev->pNext =pItem->pNext;
	}
	if(pItem->pNext){
		pItem->pNext->pPrev =pItem->pPrev;
	}
}

void CStorageImpl::ClearAllItems(){
	while(pHeader){
		ClearItem(pHeader);
		OutputDebugString("ClearAllItems");
	}
}

void CStorageImpl::ClearItem(PSTORAGE_INFO pSub){
	PSTORAGE_INFO pNext;
	if(pSub){
		pNext=pSub->pNext;
		ClearSub(pSub);
		SkipCurrentElement(pSub);
		DeleteItem(pSub);

		if(pSub==pHeader){
			pHeader=pNext;
		}
	}
}

void CStorageImpl::ClearSub(PSTORAGE_INFO pSub){
	PSTORAGE_INFO pItem;
	PSTORAGE_INFO pDel;
	pItem=pSub->pSub;
	while(pItem){
		pDel=pItem;
		pItem=pItem->pNext;
		ClearSub(pDel);
		DeleteItem(pDel);
	}
}

void CStorageImpl::DeleteItem(PSTORAGE_INFO pDel){
	OutputDebugString(pDel->szName);
	if(pDel->pLogin){
		delete pDel->pLogin;
	}
	delete pDel;
}

PSTORAGE_INFO CStorageImpl::GetHeader(){
	return pHeader;
}

void CStorageImpl::Load(wchar_t *pszFile){
	HRESULT hr;
	IStorage *pStorage;
	hr=StgOpenStorage(pszFile,NULL,STGM_READ|STGM_SHARE_EXCLUSIVE,NULL,0,&pStorage);
	if(hr==S_OK){
		DoReadInfo(pStorage,NULL);
		pStorage->Release();
	}
}

PSTORAGE_INFO CStorageImpl::DoReadInfo(IStorage* pStorage,PSTORAGE_INFO pParent){
	IEnumSTATSTG *pEnum;
	PSTORAGE_INFO pNew;
	IStorage *pSub;
	IStream *pStream;
	ULONG ulReaded;
	ULONG uFetch;
	STATSTG stg;

	pStorage->EnumElements(0,NULL,0,&pEnum);
	do{
		pEnum->Next(1,&stg,&uFetch);
		if(uFetch){
			USES_CONVERSION;
			if(stg.type ==STGTY_STORAGE){
				pNew=AddItem(W2A(stg.pwcsName),TRUE,pParent);

				pStorage->OpenStorage(stg.pwcsName,NULL,STGM_READ|STGM_SHARE_EXCLUSIVE,NULL,0,&pSub);
				DoReadInfo(pSub,pNew);
				pSub->Release();
			}else{
				pNew=AddItem(W2A(stg.pwcsName),FALSE,pParent);
				pStorage->OpenStream(stg.pwcsName,NULL,STGM_READ|STGM_SHARE_EXCLUSIVE,0,&pStream);
				if(pStream){
					pNew->pLogin =new LOGIN_INFO;
					pStream->Read((void*)pNew->pLogin,(ULONG)sizeof(LOGIN_INFO),&ulReaded);
					pStream->Release();
				}
			}
			
		}
	}while(uFetch==1);
	pEnum->Release();

	return NULL;
}

BOOL CStorageImpl::Save(wchar_t *pszFile){
	HRESULT hr;
	IStorage *pStorage;
	hr=StgCreateDocfile(pszFile,STGM_READWRITE|STGM_CREATE|STGM_SHARE_EXCLUSIVE,0,&pStorage);
	if(hr==S_OK){
		DoWriteInfo(pStorage,pHeader);
		pStorage->Commit(STGC_DEFAULT);
		pStorage->Release();
		return TRUE;
	}
	return FALSE;
}

void CStorageImpl::DoWriteInfo(IStorage *pStorage,PSTORAGE_INFO pStore){
	IStorage *pSub;
	IStream *pStream;
	ULONG ulWritten;

	while(pStore){
		USES_CONVERSION;
		if(!pStore->bDeleted){
			if(pStore->bFolder){
				pStorage->CreateStorage(A2W(pStore->szName),STGM_READWRITE|STGM_CREATE|STGM_SHARE_EXCLUSIVE,0,0,&pSub);
			}else{
				pStorage->CreateStream(A2W(pStore->szName),STGM_READWRITE|STGM_CREATE|STGM_SHARE_EXCLUSIVE,0,0,&pStream);
				pStream->Write((const void*)pStore->pLogin,(ULONG)sizeof(LOGIN_INFO),&ulWritten);
				pStream->Release();
			}
			if(pStore->pSub){
				DoWriteInfo(pSub,pStore->pSub);
				pSub->Release();
			}
		}
		pStore=pStore->pNext;
	}
}

void CStorageImpl::UpdateItemPosition(PSTORAGE_INFO pInfo,PSTORAGE_INFO pNewParent){
	PSTORAGE_INFO pNext;
	pNext=pInfo->pNext;
	DeleteRelation(pInfo);

	if(pInfo==pHeader){
		pHeader=pNext;
	}

	if(pNewParent){
		SetParent(pInfo,pNewParent);
	}else{
		InsertToRoot(pInfo);
	}
}

void CStorageImpl::SetParent(PSTORAGE_INFO pInfo,PSTORAGE_INFO pNewParent){
	pInfo->pParent =pNewParent;
	pInfo->pNext =pNewParent->pSub;
	pNewParent->pSub=pInfo;

	if(pInfo->pNext){
		pInfo->pNext->pPrev=pInfo;
	}
}

void CStorageImpl::DeleteRelation(PSTORAGE_INFO pInfo){
	if(pInfo->pParent){
		if(pInfo->pParent->pSub ==pInfo){
			pInfo->pParent->pSub =pInfo->pNext;
		}
	}
	if(pInfo->pNext){
		pInfo->pNext->pPrev =pInfo->pPrev;
	}
	if(pInfo->pPrev){
		pInfo->pPrev->pNext=pInfo->pNext;
	}
	pInfo->pParent =NULL;
	pInfo->pPrev =NULL;
	pInfo->pNext =NULL;
}

PSTORAGE_INFO CStorageImpl::UpdateItemText(PSTORAGE_INFO pInfo,const char *pszText){
	PSTORAGE_INFO pNew;

	LONG lSize;
	lSize=(long)sizeof(STORAGE_INFO)+(long)strlen(pszText)+4;
	lSize>>=2;
	lSize<<=2;
	pNew=(PSTORAGE_INFO)new char[lSize];
	memcpy((void*)pNew,(const void*)pInfo,sizeof(STORAGE_INFO));
	pNew->cb=(USHORT)lSize-sizeof(pNew->cb);

	strcpy(pNew->szName,pszText);
	if(pInfo->pParent){
		if(pInfo->pParent->pSub==pInfo){
			pInfo->pParent->pSub=pNew;
		}
	}
	PSTORAGE_INFO pSub;
	pSub=pInfo->pSub;

	while(pSub){
		pSub->pParent=pNew;
		pSub=pSub->pNext;
	}

	if(pInfo->pNext){
		pInfo->pNext->pPrev =pNew;
	}
	if(pInfo->pPrev){
		pInfo->pPrev->pNext =pNew;
	}

	delete pInfo;

	return pNew;
}

void CStorageImpl::UpdateHeader(PSTORAGE_INFO pInfo,PSTORAGE_INFO pNew){
	if(pHeader==pInfo){
		pHeader=pNew;
	}
}
